;; Generated package description from xelb.el
(define-package "xelb" "0.4" "X protocol Emacs Lisp Binding" '((emacs "24.4") (cl-generic "0.2")) :url "https://github.com/ch11ng/xelb" :keywords '("unix"))
